{% extends "basic_template.html" %}
{% block title %}
{{ super() }}
{% endblock %}
{% block head %}
{{ super() }}
{% endblock %}
{% block content %}

<h1>Source</h1>


<section class="w3-container w3-margin">
    <article>
        <aside class="w3-container w3-left">
            <center>
                <a href="https://github.com/luissv7/steadycellphenotype">
                    <img src="{{ url_for('static', filename='GitHub-Mark-64px.png') }}" alt="Github">
                </a><br>
                <a rel="license" href="http://creativecommons.org/licenses/by/4.0/">
                    <img alt="Creative Commons License" style="border-width:0"
                         src="https://i.creativecommons.org/l/by/4.0/88x31.png"/>
                </a></center>
        </aside>
        <p>
            The source code for this web site and associated utilities are licensed under the
            <a rel="license" href="http://creativecommons.org/licenses/by/4.0/">Creative Commons
                Attribution 4.0 International License</a> and is available
            on <a href="https://github.com/knappa/steadycellphenotype">Github</a>.
        </p>
        <p>
            The authors ask that anyone publishing works using, or derived from, this simulator cite
            OUR_PAPER.
        </p>
    </article>
</section>

{% endblock %}
